@@ -13,14 +13,18 @@
 #    License for the specific language governing permissions and limitations
 #    under the License.
 import contextlib
+import datetime
+import functools
 import itertools
 import os
 import ssl
 import time
 from urllib import parse
+import urllib3
 
 from oslo_log import log as logging
 from oslo_serialization import jsonutils
+import pytz
 import requests
 from requests import adapters
 
@@ -47,6 +51,7 @@ class K8sClient(object):
         token_file = config.CONF.kubernetes.token_file
         self.token = None
         self.cert = (None, None)
+        self.are_events_enabled = config.CONF.kubernetes.use_events
 
         # Setting higher numbers regarding connection pools as we're running
         # with max of 1000 green threads.
@@ -78,12 +83,16 @@ class K8sClient(object):
             else:
                 self.verify_server = ca_crt_file
 
-        self._rq_params = {
-            'cert': self.cert,
-            'verify': self.verify_server,
-            'timeout': (CONF.kubernetes.watch_connection_timeout,
-                        CONF.kubernetes.watch_read_timeout),
-        }
+        # Let's setup defaults for our Session.
+        self.session.cert = self.cert
+        self.session.verify = self.verify_server
+        if self.token:
+            self.session.headers['Authorization'] = f'Bearer {self.token}'
+        # NOTE(dulek): Seems like this is the only way to set is globally.
+        self.session.request = functools.partial(
+            self.session.request, timeout=(
+                CONF.kubernetes.watch_connection_timeout,
+                CONF.kubernetes.watch_read_timeout))
 
     def _raise_from_response(self, response):
         if response.status_code == requests.codes.not_found:
@@ -108,32 +117,58 @@ class K8sClient(object):
     def get(self, path, json=True, headers=None):
         LOG.debug("Get %(path)s", {'path': path})
         url = self._base_url + path
-        header = {}
-        if self.token:
-            header.update({'Authorization': 'Bearer %s' % self.token})
-        if headers:
-            header.update(headers)
-        response = self.session.get(url, headers=header, **self._rq_params)
+        response = self.session.get(url, headers=headers)
         self._raise_from_response(response)
-        result = response.json() if json else response.text
+
+        if json:
+            result = response.json()
+            kind = result['kind']
+
+            api_version = result.get('apiVersion')
+            if not api_version:
+                api_version = utils.get_api_ver(path)
+
+            # Strip List from e.g. PodList. For some reason `.items` of a list
+            # returned from API doesn't have `kind` set.
+            # NOTE(gryf): Also, for the sake of calculating selfLink
+            # equivalent, we need to have both: kind and apiVersion, while the
+            # latter is not present on items list for core resources, while
+            # for custom resources there are both kind and apiVersion..
+            if kind.endswith('List'):
+                kind = kind[:-4]
+
+                # NOTE(gryf): In case we get null/None for items from the API,
+                # we need to convert it to the empty list, otherwise it might
+                # be propagated to the consumers of this method and sent back
+                # to the Kubernetes as is, and fail as a result.
+                if result['items'] is None:
+                    result['items'] = []
+
+                for item in result['items']:
+                    if not item.get('kind'):
+                        item['kind'] = kind
+                    if not item.get('apiVersion'):
+                        item['apiVersion'] = api_version
+
+                if not result.get('apiVersion'):
+                    result['apiVersion'] = api_version
+        else:
+            result = response.text
+
         return result
 
     def _get_url_and_header(self, path, content_type):
         url = self._base_url + path
         header = {'Content-Type': content_type,
                   'Accept': 'application/json'}
-        if self.token:
-            header.update({'Authorization': 'Bearer %s' % self.token})
 
         return url, header
 
     def patch(self, field, path, data):
-        LOG.debug("Patch %(path)s: %(data)s", {
-            'path': path, 'data': data})
+        LOG.debug("Patch %(path)s: %(data)s", {'path': path, 'data': data})
         content_type = 'application/merge-patch+json'
         url, header = self._get_url_and_header(path, content_type)
-        response = self.session.patch(url, json={field: data},
-                                      headers=header, **self._rq_params)
+        response = self.session.patch(url, json={field: data}, headers=header)
         self._raise_from_response(response)
         return response.json().get('status')
 
@@ -159,7 +194,7 @@ class K8sClient(object):
             'path': path, 'data': data})
 
         response = self.session.patch(url, data=jsonutils.dumps(data),
-                                      headers=header, **self._rq_params)
+                                      headers=header)
         self._raise_from_response(response)
         return response.json().get('status')
 
@@ -174,7 +209,7 @@ class K8sClient(object):
                  'value': value}]
 
         response = self.session.patch(url, data=jsonutils.dumps(data),
-                                      headers=header, **self._rq_params)
+                                      headers=header)
         self._raise_from_response(response)
         return response.json().get('status')
 
@@ -187,7 +222,7 @@ class K8sClient(object):
                  'path': '/metadata/annotations/{}'.format(annotation_name)}]
 
         response = self.session.patch(url, data=jsonutils.dumps(data),
-                                      headers=header, **self._rq_params)
+                                      headers=header)
         self._raise_from_response(response)
         return response.json().get('status')
 
@@ -206,7 +241,7 @@ class K8sClient(object):
         data = [{'op': 'remove',
                  'path': f'/metadata/annotations/{annotation_name}'}]
         response = self.session.patch(url, data=jsonutils.dumps(data),
-                                      headers=header, **self._rq_params)
+                                      headers=header)
         if response.ok:
             return response.json().get('status')
         raise exc.K8sClientException(response.text)
@@ -215,11 +250,8 @@ class K8sClient(object):
         LOG.debug("Post %(path)s: %(body)s", {'path': path, 'body': body})
         url = self._base_url + path
         header = {'Content-Type': 'application/json'}
-        if self.token:
-            header.update({'Authorization': 'Bearer %s' % self.token})
 
-        response = self.session.post(url, json=body, headers=header,
-                                     **self._rq_params)
+        response = self.session.post(url, json=body, headers=header)
         self._raise_from_response(response)
         return response.json()
 
@@ -227,10 +259,8 @@ class K8sClient(object):
         LOG.debug("Delete %(path)s", {'path': path})
         url = self._base_url + path
         header = {'Content-Type': 'application/json'}
-        if self.token:
-            header.update({'Authorization': 'Bearer %s' % self.token})
 
-        response = self.session.delete(url, headers=header, **self._rq_params)
+        response = self.session.delete(url, headers=header)
         self._raise_from_response(response)
         return response.json()
 
@@ -240,7 +270,7 @@ class K8sClient(object):
         if finalizer in obj['metadata'].get('finalizers', []):
             return True
 
-        path = obj['metadata']['selfLink']
+        path = utils.get_res_link(obj)
         LOG.debug(f"Add finalizer {finalizer} to {path}")
         url, headers = self._get_url_and_header(
             path, 'application/merge-patch+json')
@@ -256,8 +286,7 @@ class K8sClient(object):
                 },
             }
 
-            response = self.session.patch(url, json=data, headers=headers,
-                                          **self._rq_params)
+            response = self.session.patch(url, json=data, headers=headers)
 
             if response.ok:
                 return True
@@ -268,7 +297,11 @@ class K8sClient(object):
                 # Object is being deleting or gone. Return.
                 return False
             except exc.K8sConflict:
-                obj = self.get(path)
+                try:
+                    obj = self.get(path)
+                except exc.K8sResourceNotFound:
+                    # Object got removed before finalizer was set
+                    return False
                 if finalizer in obj['metadata'].get('finalizers', []):
                     # Finalizer is there, return.
                     return True
@@ -277,7 +310,7 @@ class K8sClient(object):
         self._raise_from_response(response)
 
     def remove_finalizer(self, obj, finalizer):
-        path = obj['metadata']['selfLink']
+        path = utils.get_res_link(obj)
         LOG.debug(f"Remove finalizer {finalizer} from {path}")
         url, headers = self._get_url_and_header(
             path, 'application/merge-patch+json')
@@ -297,8 +330,7 @@ class K8sClient(object):
                 },
             }
 
-            response = self.session.patch(url, json=data, headers=headers,
-                                          **self._rq_params)
+            response = self.session.patch(url, json=data, headers=headers)
 
             if response.ok:
                 return True
@@ -349,8 +381,7 @@ class K8sClient(object):
             if resource_version:
                 metadata['resourceVersion'] = resource_version
             data = jsonutils.dumps({"metadata": metadata}, sort_keys=True)
-            response = self.session.patch(url, data=data,
-                                          headers=header, **self._rq_params)
+            response = self.session.patch(url, data=data, headers=header)
             if response.ok:
                 return response.json()['metadata'].get('annotations', {})
             if response.status_code == requests.codes.conflict:
@@ -381,9 +412,6 @@ class K8sClient(object):
     def watch(self, path):
         url = self._base_url + path
         resource_version = None
-        header = {}
-        if self.token:
-            header.update({'Authorization': 'Bearer %s' % self.token})
 
         attempt = 0
         while True:
@@ -393,8 +421,7 @@ class K8sClient(object):
                     params['resourceVersion'] = resource_version
                 with contextlib.closing(
                         self.session.get(
-                            url, params=params, stream=True, headers=header,
-                            **self._rq_params)) as response:
+                            url, params=params, stream=True)) as response:
                     if not response.ok:
                         raise exc.K8sClientException(response.text)
                     attempt = 0
@@ -408,7 +435,8 @@ class K8sClient(object):
                             m = line_dict.get('object', {}).get('metadata', {})
                             resource_version = m.get('resourceVersion', None)
             except (requests.ReadTimeout, requests.ConnectionError,
-                    ssl.SSLError, requests.exceptions.ChunkedEncodingError):
+                    ssl.SSLError, requests.exceptions.ChunkedEncodingError,
+                    urllib3.exceptions.SSLError):
                 t = utils.exponential_backoff(attempt)
                 log = LOG.debug
                 if attempt > 0:
@@ -420,3 +448,68 @@ class K8sClient(object):
                     params.get('resourceVersion'))
                 time.sleep(t)
                 attempt += 1
+
+    def add_event(self, resource, reason, message, type_='Normal',
+                  component='kuryr-controller'):
+        """Create an Event object for the provided resource."""
+        if not self.are_events_enabled:
+            return {}
+
+        if not resource:
+            return {}
+
+        involved_object = {'apiVersion': resource['apiVersion'],
+                           'kind': resource['kind'],
+                           'name': resource['metadata']['name'],
+                           'namespace': resource['metadata']['namespace'],
+                           'uid': resource['metadata']['uid']}
+
+        # This is needed for Event date, otherwise LAST SEEN/Age will be empty
+        # and misleading.
+        now = datetime.datetime.utcnow().replace(tzinfo=pytz.UTC)
+        date_time = now.strftime("%Y-%m-%dT%H:%M:%SZ")
+
+        name = ".".join((resource['metadata']['name'],
+                         self._get_hex_timestamp(now)))
+
+        event = {'kind': 'Event',
+                 'apiVersion': 'v1',
+                 'firstTimestamp': date_time,
+                 'metadata': {'name': name},
+                 'reason': reason,
+                 'message': message,
+                 'type': type_,
+                 'involvedObject': involved_object,
+                 'source': {'component': component,
+                            'host': utils.get_nodename()}}
+
+        try:
+            return self.post(f'{constants.K8S_API_BASE}/namespaces/'
+                             f'{resource["metadata"]["namespace"]}/events',
+                             event)
+        except exc.K8sNamespaceTerminating:
+            # We can't create events in a Namespace that is being terminated,
+            # there's no workaround, no need to log it, just ignore it.
+            return {}
+        except exc.K8sClientException:
+            LOG.warning(f'There was non critical error during creating an '
+                        'Event for resource: "{resource}", with reason: '
+                        f'"{reason}", message: "{message}" and type: '
+                        f'"{type_}"')
+            return {}
+
+    def _get_hex_timestamp(self, datetimeobj):
+        """Get hex representation for timestamp.
+
+        In Kuberenets, Event name is constructed name of the bounded object
+        and timestamp in hexadecimal representation.
+        Note, that Python timestamp is represented as floating figure:
+          1631622163.8534190654754638671875
+        while those which origin from K8s, after change to int:
+          1631622163915909162
+        so, to get similar integer, we need to multiply the float by
+        100000000 to get the same precision and cast to integer, to get rid
+        of the fractures, and finally convert it to hex representation.
+        """
+        timestamp = datetime.datetime.timestamp(datetimeobj)
+        return format(int(timestamp * 100000000), 'x')
