@@ -13,8 +13,9 @@
 #    License for the specific language governing permissions and limitations
 #    under the License.
 
+from unittest import mock
+
 import fixtures
-import mock
 
 from kuryr_kubernetes import k8s_client
 
@@ -27,17 +28,25 @@ class MockK8sClient(fixtures.Fixture):
             lambda: self.client))
 
 
-class MockNeutronClient(fixtures.Fixture):
-    def _setUp(self):
-        self.client = mock.Mock()
-        self.useFixture(fixtures.MockPatch(
-            'kuryr_kubernetes.clients.get_neutron_client',
-            lambda: self.client))
-
-
 class MockLBaaSClient(fixtures.Fixture):
     def _setUp(self):
         self.client = mock.Mock()
         self.useFixture(fixtures.MockPatch(
             'kuryr_kubernetes.clients.get_loadbalancer_client',
             lambda: self.client))
+
+
+class MockNetworkClient(fixtures.Fixture):
+    def _setUp(self):
+        self.client = mock.Mock()
+        self.useFixture(fixtures.MockPatch(
+            'kuryr_kubernetes.clients.get_network_client',
+            lambda: self.client))
+
+
+class MockComputeClient(fixtures.Fixture):
+    def _setUp(self):
+        self.client = mock.Mock()
+        self.useFixture(fixtures.MockPatch(
+            'kuryr_kubernetes.clients.get_compute_client',
+            lambda: self.client))
